#include <gtest/gtest.h>

#include <structs.h>
#include <utils.h>

#define PIECE_A_OLD	"   "
#define PIECE_A_NEW	"   "
#define PIECE_B_OLD	""
#define PIECE_B_NEW	"bla-bla-bla"
#define PIECE_C_OLD	"\n "
#define PIECE_C_NEW	"bla-bla-bla"
#define PIECE_D_OLD	"."
#define PIECE_D_NEW	"tail"
#define SAMPLE_PIECE PIECE_A_OLD "    ,   .   ,\n" \
	" " PIECE_B_OLD "    ,  \n" \
	"  -    - ,   , " PIECE_C_OLD \
	"    ,    \n" \
	" .   -   ,   - " PIECE_D_OLD
#define SAMPLE_PIECE_R1 PIECE_A_NEW "    ,   .   ,\n" \
	" " PIECE_B_OLD "    ,  \n" \
	"  -    - ,   , " PIECE_C_OLD \
	"    ,    \n" \
	" .   -   ,   - " PIECE_D_OLD
#define SAMPLE_PIECE_R2 PIECE_A_NEW "    ,   .   ,\n" \
	" " PIECE_B_NEW "    ,  \n" \
	"  -    - ,   , " PIECE_C_OLD \
	"    ,    \n" \
	" .   -   ,   - " PIECE_D_OLD
#define SAMPLE_PIECE_R3 PIECE_A_NEW "    ,   .   ,\n" \
	" " PIECE_B_NEW "    ,  \n" \
	"  -    - ,   , " PIECE_C_NEW \
	"    ,    \n" \
	" .   -   ,   - " PIECE_D_OLD
#define SAMPLE_PIECE_R4 PIECE_A_NEW "    ,   .   ,\n" \
	" " PIECE_B_NEW "    ,  \n" \
	"  -    - ,   , " PIECE_C_NEW \
	"    ,    \n" \
	" .   -   ,   - " PIECE_D_NEW

#define SAMPLE_NUMBERS_PIECE "123456789012345678901"
#define SAMPLE_NUMBERS_PATTERN_PIECE_FROM_1 "1"
#define SAMPLE_NUMBERS_PATTERN_PIECE_TO_1 "a"
#define SAMPLE_NUMBERS_PIECE_R1 "a234567890a234567890a"
#define SAMPLE_NUMBERS_PATTERN_PIECE_FROM_2 "2"
#define SAMPLE_NUMBERS_PATTERN_PIECE_TO_2 "bbbb"

const char* SAMPLE_STRING = SAMPLE_PIECE;
const char* SAMPLE_STRING_R1 = SAMPLE_PIECE_R1;
const char* SAMPLE_STRING_R2 = SAMPLE_PIECE_R2;
const char* SAMPLE_STRING_R3 = SAMPLE_PIECE_R3;
const char* SAMPLE_STRING_R4 = SAMPLE_PIECE_R4;
const char* DOUBLE_SAMPLE_STRING = SAMPLE_PIECE SAMPLE_PIECE;

const char* SAMPLE_NUMBERS_STRING = SAMPLE_NUMBERS_PIECE;

const char* SAMPLE_NUMBERS_PATTERN_FROM_1 = SAMPLE_NUMBERS_PATTERN_PIECE_FROM_1;
const char* SAMPLE_NUMBERS_PATTERN_TO_1 = SAMPLE_NUMBERS_PATTERN_PIECE_TO_1;
const int SAMPLE_NUMBERS_SUBSTRINGS_COUNT = 3;
const char* SAMPLE_NUMBERS_STRING_R1 = SAMPLE_NUMBERS_PIECE_R1;

const char* SAMPLE_NUMBERS_PATTERN_FROM_2 = SAMPLE_NUMBERS_PATTERN_PIECE_FROM_2;
const char* SAMPLE_NUMBERS_PATTERN_TO_2 = SAMPLE_NUMBERS_PATTERN_PIECE_TO_2;
const char* SAMPLE_NUMBERS_STRING_R2 = SAMPLE_NUMBERS_PIECE;

AbstractStringWriter::shared_ptr create_delegated_string_writer(char*& delegated_string)
{
	const auto result = AbstractStringWriter::shared_ptr(new DelegatedStringWriter(delegated_string));
	return result;
}

AbstractStringWriter::shared_ptr create_std_string_writer()
{
	const auto result = AbstractStringWriter::shared_ptr(new StdStringWriter());
	return result;
}

TEST(Utils_Editor_DelegatedStringWriter, Set)
{
	char* string = nullptr;
	const auto writer = create_delegated_string_writer(string);
	writer->set_string(SAMPLE_STRING);

	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_STRING));
	EXPECT_EQ(0, strcmp(string, SAMPLE_STRING));
	free(string);
}

TEST(Utils_Editor_StdStringWriter, Set)
{
	const auto writer = create_std_string_writer();
	writer->set_string(SAMPLE_STRING);

	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_STRING));
}

TEST(Utils_Editor_DelegatedStringWriter, AppendToEmpty)
{
	char* string = nullptr;
	const auto writer = create_delegated_string_writer(string);
	writer->append_string(SAMPLE_STRING);

	const auto equality = strcmp(writer->get_string(), SAMPLE_STRING);
	EXPECT_EQ(0, equality);
	free(string);
}

TEST(Utils_Editor_StdStringWriter, AppendToEmpty)
{
	const auto writer = create_std_string_writer();
	writer->append_string(SAMPLE_STRING);

	const auto equality = strcmp(writer->get_string(), SAMPLE_STRING);
	EXPECT_EQ(0, equality);
}

TEST(Utils_Editor_DelegatedStringWriter, Append)
{
	char* string = nullptr;
	const auto writer = create_delegated_string_writer(string);
	writer->set_string(SAMPLE_STRING);
	writer->append_string(SAMPLE_STRING);

	EXPECT_EQ(0, strcmp(writer->get_string(), DOUBLE_SAMPLE_STRING));
	free(string);
}

TEST(Utils_Editor_StdStringWriter, Append)
{
	const auto writer = create_std_string_writer();
	writer->set_string(SAMPLE_STRING);
	writer->append_string(SAMPLE_STRING);

	EXPECT_EQ(0, strcmp(writer->get_string(), DOUBLE_SAMPLE_STRING));
}

TEST(Utils_Editor_DelegatedStringWriter, Length)
{
	char* string = nullptr;
	const auto writer = create_delegated_string_writer(string);
	writer->set_string(SAMPLE_STRING);

	EXPECT_EQ(strlen(SAMPLE_STRING), writer->length());
	free(string);
}

TEST(Utils_Editor_StdStringWriter, Length)
{
	const auto writer = create_std_string_writer();
	writer->set_string(SAMPLE_STRING);

	EXPECT_EQ(strlen(SAMPLE_STRING), writer->length());
}

TEST(Utils_Editor_DelegatedStringWriter, ReplaceOne)
{
	char* string = nullptr;
	const auto writer = create_delegated_string_writer(string);
	writer->set_string(SAMPLE_STRING);

	const auto replace_result1 = replace_str(writer, PIECE_A_OLD, PIECE_A_NEW, 0, MAX_STRING_LENGTH);
	EXPECT_EQ(1, replace_result1);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_PIECE_R1));

	const auto replace_result2 = replace_str(writer, PIECE_B_OLD, PIECE_B_NEW, 0, MAX_STRING_LENGTH);
	EXPECT_EQ(1, replace_result2);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_PIECE_R2));

	const auto replace_result3 = replace_str(writer, PIECE_C_OLD, PIECE_C_NEW, 0, MAX_STRING_LENGTH);
	EXPECT_EQ(1, replace_result3);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_PIECE_R3));

	const auto replace_result4 = replace_str(writer, PIECE_D_OLD, PIECE_D_NEW, 0, MAX_STRING_LENGTH);
	EXPECT_EQ(1, replace_result4);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_PIECE_R4));
	free(string);
}

TEST(Utils_Editor_StdStringWriter, ReplaceOne)
{
	const auto writer = create_std_string_writer();
	writer->set_string(SAMPLE_STRING);

	const auto replace_result1 = replace_str(writer, PIECE_A_OLD, PIECE_A_NEW, 0, MAX_STRING_LENGTH);
	EXPECT_EQ(1, replace_result1);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_PIECE_R1));

	const auto replace_result2 = replace_str(writer, PIECE_B_OLD, PIECE_B_NEW, 0, MAX_STRING_LENGTH);
	EXPECT_EQ(1, replace_result2);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_PIECE_R2));

	const auto replace_result3 = replace_str(writer, PIECE_C_OLD, PIECE_C_NEW, 0, MAX_STRING_LENGTH);
	EXPECT_EQ(1, replace_result3);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_PIECE_R3));

	const auto replace_result4 = replace_str(writer, PIECE_D_OLD, PIECE_D_NEW, 0, MAX_STRING_LENGTH);
	EXPECT_EQ(1, replace_result4);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_PIECE_R4));
}

TEST(Utils_Editor_DelegatedStringWriter, ReplaceMany)
{
	char* string = nullptr;
	const auto writer = create_delegated_string_writer(string);
	writer->set_string(SAMPLE_NUMBERS_STRING);
	const auto result = replace_str(writer, SAMPLE_NUMBERS_PATTERN_FROM_1, SAMPLE_NUMBERS_PATTERN_TO_1, 1, MAX_STRING_LENGTH);

	EXPECT_EQ(SAMPLE_NUMBERS_SUBSTRINGS_COUNT, result);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_NUMBERS_PIECE_R1));
	free(string);
}

TEST(Utils_Editor_StdStringWriter, ReplaceMany)
{
	const auto writer = create_std_string_writer();
	writer->set_string(SAMPLE_NUMBERS_STRING);
	const auto result = replace_str(writer, SAMPLE_NUMBERS_PATTERN_FROM_1, SAMPLE_NUMBERS_PATTERN_TO_1, 1, MAX_STRING_LENGTH);

	EXPECT_EQ(SAMPLE_NUMBERS_SUBSTRINGS_COUNT, result);
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_NUMBERS_PIECE_R1));
}

TEST(Utils_Editor_DelegatedStringWriter, ReplaceManyOverflow)
{
	char* string = nullptr;
	const auto writer = create_delegated_string_writer(string);
	writer->set_string(SAMPLE_NUMBERS_STRING);

	const auto expected_result_length = strlen(SAMPLE_NUMBERS_PIECE);
	const auto result = replace_str(writer, SAMPLE_NUMBERS_PATTERN_FROM_2, SAMPLE_NUMBERS_PATTERN_TO_2, 1, static_cast<int>(expected_result_length));
	EXPECT_EQ(-1, result);
	EXPECT_EQ(expected_result_length, writer->length());
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_NUMBERS_PIECE));
	free(string);
}

TEST(Utils_Editor_StdStringWriter, ReplaceManyOverflow)
{
	const auto writer = create_std_string_writer();
	writer->set_string(SAMPLE_NUMBERS_STRING);

	const auto expected_result_length = strlen(SAMPLE_NUMBERS_PIECE);
	const auto result = replace_str(writer, SAMPLE_NUMBERS_PATTERN_FROM_2, SAMPLE_NUMBERS_PATTERN_TO_2, 1, static_cast<int>(expected_result_length));
	EXPECT_EQ(-1, result);
	EXPECT_EQ(expected_result_length, writer->length());
	EXPECT_EQ(0, strcmp(writer->get_string(), SAMPLE_NUMBERS_PIECE));
}

// vim: ts=4 sw=4 tw=0 noet syntax=cpp :
